; threads v1.0, functions for cooperative multithreading
; Copyright 2002, 2003 Chris Williams <cow3.14@juno.com>
; 
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
; 
; Started       2002 Jan 09
; Last modified 2003 Jun 26

; To include optional function foo, for example, define NEED_thread_foo.
; Optional functions: setstate, cancel, suspend, join.

STATE_RUNNING	= 0
STATE_SUSPENDED	= 1
STATE_STARTED	= 2
STATE_EXITED	= 3
STATE_CANCELLED	= 4

#ifdef NEED_thread_join
sizeof_thread_t		= 11
#else
sizeof_thread_t		= 9
#endif
thread_t.node		= 0
thread_t.node.next	= 0
thread_t.code		= 2
thread_t.sp		= 4
thread_t.stack_top	= 6
thread_t.state		= 8
thread_t.retval		= 9
	
sizeof_list_t	= 4
list_t.head	= 0
list_t.tail	= 2

sizeof_node_t	= 2
node_t.next	= 0

;#define SIM_ONLY ; don't change the real stack pointer (SP)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; push_all and pop_all
; pushes or pops all registers except IX and IY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#define push_all push af\ push bc\ push de\ push hl
#define pop_all pop hl\ pop de\ pop bc\ pop af
	
.module thread
	ld	hl,0
	ld	(_threads+list_t.head),hl
	ld	(_threads+list_t.tail),hl	; threads.head = threads.tail = NULL
	add	hl,sp
	ld	(_main_thread+thread_t.sp),hl
	ld	(_main_thread+thread_t.stack_top),hl	; main_thread.sp = main_thread.stack_top = sp
#if STATE_RUNNING == 0
	xor	a
#else
	ld	a,STATE_RUNNING
#endif
	ld	(_main_thread+thread_t.state),a	; main_thread.state = STATE_RUNNING
	ld	hl,_threads
	ld	de,_main_thread
	call	list_append_node	; list_append_node(&threads, (node_t *)&main_thread)
	ld	a,1
	ld	(_num_running_threads),a
	jp	_end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread to create
; 	DE	address of code
; 	BC	top address of stack
; output:
; 	A = 0 on success, !0 on error
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_create:
	ld	a,l
	or	h			; t == NULL
	ld	a,-1
	ret	z			; return -1
#if thread_t.code != 0
#if thread_t.code == 1
	inc	hl
#else
		push	de
	ld	de,thread_t.code
	add	hl,de
		pop	de
#endif
#endif
	ld	(hl),e
	inc	hl
	ld	(hl),d			; t->code = code
#if thread_t.stack_top-thread_t.code == 2
	inc	hl
#else
	ld	de,thread_t.stack_top-thread_t.code-1
	add	hl,de
#endif
	ld	(hl),c
	inc	hl
	ld	(hl),b			; t->stack_top = stack_top
#if thread_t.state-thread_t.stack_top == 2
	inc	hl
#else
	ld	de,thread_t.state-thread_t.stack_top-1
	add	hl,de
#endif
	ld	(hl),STATE_STARTED	; t->state = STATE_STARTED
#if thread_t.node-thread_t.state == 1
	inc	hl
#else
	ld	de,thread_t.node-thread_t.state
	add	hl,de
#endif
	ex	de,hl
	ld	hl,_threads
	call	list_append_node	; list_append_node(&threads, &t->node)
	;call	thread_cooperate	; thread_cooperate()
	xor	a
	ret				; return 0

#ifdef NEED_thread_setstate
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread
; 	A	new state for thread
; output:
; 	A = 0 on success, !0 on error
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_setstate:
#if STATE_RUNNING == 0
	or	a
#else
	cp	STATE_RUNNING
#endif
	jr	z,_setstate0
#if STATE_SUSPENDED == 0
	or	a
#else
	cp	STATE_SUSPENDED
#endif
	jr	z,_setstate0	; state != STATE_RUNNING && state != STATE_SUSPENDED
	ld	a,-1
	ret			; return -1
_setstate0
#if thread_t.state != 0
	ld	bc,thread_t.state
	add	hl,bc
#endif
	cp	(hl)
	jr	nz,_setstate1	; t->state == state
	xor	a
	ret			; return 0
_setstate1
		push	hl
	ld	hl,_num_running_threads
#if STATE_RUNNING == 0
	or	a
#else
	cp	STATE_RUNNING
#endif
	jr	nz,_setstate2	; state == STATE_RUNNING
	inc	(hl)		; ++num_running_threads
	jr	_setstate3
_setstate2
	ld	b,a
	ld	a,(hl)
#if STATE_RUNNING == 0
	or	a
#else
	cp	STATE_RUNNING
#endif
	ld	a,b
	jr	nz,_setstate3	; t->state == STATE_RUNNING
	dec	(hl)		; --num_running_threads
_setstate3
		pop	hl
	ld	(hl),a		; t->state = state
	xor	a
	ret			; return 0
#endif

#ifdef NEED_thread_cancel
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread to cancel
; output:
; 	A = 0 on success, !0 on error
; registers destroyed:
; 	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_cancel:
	ld	a,l
	or	h			; t == NULL
	ld	a,-1
	ret	z			; return -1
	ld	de,_main_thread
	ld	a,h
	cp	d
	jr	nz,_cancel1
	ld	a,l
	cp	e			; t == &main_thread
	jr	nz,_cancel1
#if thread_t.stack_top != 0
	ld	hl,thread_t.stack_top
	add	hl,de
#else
	ex	de,hl
#endif
	call	LD_HL_MHL
#ifdef SIM_ONLY
	ld	(_sp),hl
#else
	ld	sp,hl			; sp = main_thread.stack_top
#endif
	ret				; return 0
_cancel1
#if thread_t.state != 0
	ld	de,thread_t.state
	add	hl,de
#endif
	ld	a,(hl)
#if STATE_RUNNING == 0
	or	a
#else
	cp	STATE_RUNNING
#endif
	jr	nz,_cancel2
	ex	de,hl
	ld	hl,_num_running_threads
	dec	(hl)
	ex	de,hl			; --num_running_threads
_cancel2
	ld	(hl),STATE_CANCELLED	; t->state = STATE_CANCELLED
	call	thread_cooperate
	xor	a			; return 0
	ret
#endif

#ifdef NEED_thread_suspend
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	none
; output:
; 	none
; registers destroyed:
; 	none
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_suspend:
		push	af
		push	hl
		push	de
	ld	hl,(_threads+list_t.head)
#if thread_t.state != 0
	ld	de,thread_t.state
	add	hl,de			; s = &((thread_t *)threads.head)->state
#endif
	ld	a,(hl)
#if STATE_SUSPENDED == 0
	or	a
#else
	cp	STATE_SUSPENDED		; if (*s == STATE_SUSPENDED)
#endif
	jr	nz,_suspend1
	call	thread_cooperate	; thread_cooperate()
	jr	_suspend2
_suspend1				; else
	ld	(hl),STATE_SUSPENDED	; *s = STATE_SUSPENDED
	ex	de,hl
	ld	hl,_num_running_threads
	dec	(hl)			; --num_running_threads
	call	thread_cooperate	; thread_cooperate()
	inc	(hl)			; ++num_running_threads
	ex	de,hl
	ld	(hl),STATE_RUNNING
_suspend2
		pop	de
		pop	hl
		pop	af
	ret				; return
#endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	none
; output:
; 	none
; registers destroyed:
; 	none
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_cooperate:
		push_all
#ifdef SIM_ONLY
	ld	a,(_threads+list_t.head+1)
	call	puthex
	ld	a,(_threads+list_t.head)
	call	puthex
	ld	a,' '
	call	putc
	ld	a,(_sp+1)
	call	puthex
	ld	a,(_sp)
	call	puthex
#endif
#ifdef SIM_ONLY
	ld	de,(_sp)
#else
	ld	hl,0
	add	hl,sp
	ex	de,hl
#endif
	ld	hl,(_threads+list_t.head)
	ld	(_threads+list_t.tail),hl	; threads.tail = threads.head
#if thread_t.sp != 0
	ld	bc,thread_t.sp
	add	hl,bc
#endif
	ld	(hl),e
	inc	hl
	ld	(hl),d			; ((thread_t *)threads.head)->sp = sp
;	ld	hl,(_threads+list_t.head)
	ld	bc,thread_t.node.next-thread_t.sp-1
	add	hl,bc
	call	LD_HL_MHL
	ld	(_threads+list_t.head),hl	; threads.head = threads.head->next
_cooperate1				; while (1) {
	ld	hl,(_threads+list_t.head)
#if thread_t.state != 0
	ld	bc,thread_t.state
	add	hl,bc
#endif
	ld	a,(hl)			; switch (((thread_t *)threads.head)->state) {
#if STATE_RUNNING == 0
	or	a
#else
	cp	STATE_RUNNING		; case STATE_RUNNING
#endif
	jr	nz,_cooperate2
_running	; HL must be (_threads+list_t.head)+thread_t.state
	ld	bc,thread_t.sp-thread_t.state
	add	hl,bc
	call	LD_HL_MHL
#ifdef SIM_ONLY
	ld	(_sp),hl
#else
	ld	sp,hl			; sp = ((thread_t *)threads.head)->sp
#endif
		pop_all
	ret	; return to a running thread
_cooperate2
#if STATE_SUSPENDED == 0
	or	a
#else
	cp	STATE_SUSPENDED		; case STATE_SUSPENDED
#endif
	jr	nz,_cooperate3
	ld	de,(_threads+list_t.head)
	ld	a,e
	cp	_main_thread&255
	jr	nz,_running
	ld	a,d
	cp	_main_thread>>8		; (thread_t *)threads.head = &main_thread
	jr	nz,_running
	ld	a,(_num_running_threads)
	or	a			; num_running_threads == 0
	jr	nz,_running
	halt
	jr	_running		; goto running
_cooperate3
#if STATE_CANCELLED == 0
	or	a
#else
	cp	STATE_CANCELLED		; case STATE_CANCELLED
#endif
	jr	nz,_cooperate4
	jr	thread_exit		; thread_exit(0)
_cooperate4
#if STATE_STARTED == 0
	or	a
#else
	cp	STATE_STARTED		; case STATE_STARTED
#endif
	jr	nz,_cooperate5
	ld	hl,(_threads+list_t.head)
#if thread_t.stack_top != 0
	ld	bc,thread_t.stack_top
	add	hl,bc
#endif
	ld	b,h
	ld	c,l			; t = (thread_t *)threads.head
	call	LD_HL_MHL
#ifdef SIM_ONLY
	ld	(_sp),hl
#else
	ld	sp,hl			; sp = t->stack_top
#endif
	ld	hl,_num_running_threads
	inc	(hl)			; ++num_running_threads
	ld	hl,thread_t.state-thread_t.stack_top
	add	hl,bc
		push	hl
	ld	(hl),STATE_RUNNING	; t->state = STATE_RUNNING
	ld	bc,thread_t.code-thread_t.state
	add	hl,bc
#ifdef NEED_thread_join
		push	hl
#endif
	call	LD_HL_MHL
#ifdef SIM_ONLY
	call	thread_cooperate
#else
	ld	bc,$+5
	push	bc			; return address
	jp	(hl)			; t->code()
#endif
#ifdef NEED_thread_join
		pop	bc
	ex	de,hl
	ld	hl,thread_t.retval-thread_t.code
	add	hl,bc
	ld	(hl),e
	inc	hl
	ld	(hl),d			; t->retval = ...
#endif
		pop	hl		; hl = &t->state
	ld	a,(hl)
#if STATE_RUNNING == 0
	or	a
#else
	cp	STATE_RUNNING		; t->state == STATE_RUNNING
#endif
	jr	nz,_cooperate6
		push	hl
	ld	hl,_num_running_threads
	dec	(hl)			; --num_running_threads
		pop	hl
_cooperate6
#if STATE_CANCELLED == 0
	or	a
#else
	cp	STATE_CANCELLED		; t->state != STATE_CANCELLED
#endif
	jr	z,_cooperate7
	ld	(hl),STATE_EXITED	; t->state = STATE_EXITED
_cooperate7
	ld	hl,_threads
_cooperate5				; default
	call	list_remove_head	; list_remove_head(&threads)
	jr	_cooperate1

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	return value
; output:
; 	none (does not return)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_exit:
#ifdef NEED_thread_join
	ex	de,hl
#endif
	ld	hl,(_threads+list_t.head)
	ld	bc,thread_t.stack_top
	add	hl,bc
	call	LD_HL_MHL
	ld	bc,-4
	add	hl,bc
#ifdef SIM_ONLY
	ld	(_sp),hl
#else
	ld	sp,hl			; sp = ((thread_t *)threads.head)->stack_top
#endif
#ifdef NEED_thread_join
	ex	de,hl			; return retval
#endif
	ret

#ifdef NEED_thread_join
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread to join
; output:
; 	A = 0 if the thread exitted
; 	A != 0 if the thread was cancelled
; 	HL = thread's return value (only if A == 0)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_join:
#if thread_t.state != 0
	ld	de,thread_t.state
	add	hl,de
#endif
		push	hl
_join1
	ld	a,(hl)
	cp	STATE_CANCELLED		; t->state != STATE_CANCELLED
	jr	z,_join2
	cp	STATE_EXITED		; t->state != STATE_EXITED
	call	nz,thread_cooperate	; thread_cooperate()
	jr	nz,_join1
		pop	hl
	call	LD_HL_MHL
	xor	a
	ret				; return t->retval
_join2
	ld	a,1
		pop	hl
	ret
#endif

_threads				; list_t threads
	.ds	sizeof_list_t
_main_thread:				; thread_t main_thread
	.ds	sizeof_thread_t
_num_running_threads:			; int num_running_threads
	.db	0
#ifdef SIM_ONLY
_sp:	.dw 0
#endif

.module list
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of list
; 	DE	address of node
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
list_append_node:
	ld	a,l
	or	h
	ld	a,-1
	ret	z			; list == NULL, return -1
		push	hl
	call	LD_HL_MHL
	ld	a,l
	or	h
	jr	nz,_append_node1	; list->head == NULL
	ld	h,d
	ld	l,e
	ld	(hl),e
	inc	hl
	ld	(hl),d			; node->next = node
		pop	hl
		push	hl
	ld	(hl),e
	inc	hl
	ld	(hl),d			; list->head = node
	jr	_append_node2
_append_node1				; hl = list->head, de = node
	ex	de,hl
	ld	(hl),e
	inc	hl
	ld	(hl),d			; node->next = list->head
	dec	hl
	ex	de,hl
		pop	hl		; hl = list
		push	hl
	inc	hl
	inc	hl
	call	LD_HL_MHL		; hl = &list->tail->next
	ld	(hl),e
	inc	hl
	ld	(hl),d			; list->tail->next = node
_append_node2
		pop	hl
	inc	hl
	inc	hl
	ld	(hl),e
	inc	hl
	ld	(hl),d			; list->tail = node
	xor	a
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of list
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
list_remove_head:
	ld	a,l
	or	h
	ld	a,-1
	ret	z			; list == NULL
	ld	a,(hl)
	inc	hl
	or	(hl)
	dec	hl
	ld	a,-1
	ret	z			; list->head == NULL
		push	hl
	ld	b,(hl)
	inc	hl
	ld	a,(hl)
	inc	hl
	cp	(hl)
	jr	nz,_remove_head1
	inc	hl
	ld	a,(hl)
	cp	b
	jr	nz,_remove_head1	; list->head == list->tail
		pop	hl
	ld	(hl),a
	inc	hl
	ld	(hl),a			; list->head = NULL
	inc	hl
	ld	(hl),a
	inc	hl
	ld	(hl),a			; list->tail = NULL
	jr	_remove_head2
_remove_head1
		pop	hl
		push	hl
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
		pop	hl
		push	de
	ld	(hl),e
	inc	hl
	ld	(hl),d			; list->head = list->head->next
	inc	hl
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl
		pop	de
	ld	(hl),e
	inc	hl
	ld	(hl),d			; list->tail->next = list->head->next
_remove_head2
	xor	a
	ret
.module noname

#ifdef SIM_ONLY
putc = _putc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	A	nibble to put in hexadecimal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
puthex:
	push	af
	srl	a
	srl	a
	srl	a
	srl	a
	call	puthexdigit
	pop	af
	and	$0F
	; fall through

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	A	nibble to put in hexadecimal
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
puthexdigit:
	add	a,'0'
	cp	'9'+1
	jr	c,$+4
	add	a,'A'-'0'-10

	jp	putc
#endif

.module thread
_end
.module noname
