/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef _THREADS_H_
#define _THREADS_H_

#include "lists.h"

#ifndef NULL
# define NULL 0
#endif

/* Processor's own stack pointer. It's not an actual
 * variable and can be changed only with assembly. */
extern void *sp;

typedef struct thread_t {
	node_t node;
	void *(*code)(void);
	void *sp;
	void *stack_top;
	enum thread_state {STATE_RUNNING, STATE_SUSPENDED, STATE_STARTED, STATE_EXITED, STATE_CANCELLED} state;
	void *retval;
} thread_t;

void threads_init(void);
int thread_create(thread_t *t, void *(*code)(void), void *stack_top);
int thread_setstate(thread_t *t, enum thread_state state);
int thread_cancel(thread_t *t);
void thread_suspend(void);
void thread_cooperate(void);
void *thread_exit(void *retval);
void *thread_join(thread_t *t);

#endif /* _THREADS_H_ */
